/** edge recombination
* Kreuzung zweier Individuen mit Hilfe des Verfahrens ERX
*/

import java.util.*;
public class ERX implements IKreuzung{

 /** Konstruktor, Festlegung der Ausgangssituation */
 public ERX(){
    kanten=null;
 }

 /**
* Implementierung der geforderten Methode namens "kreuzung".
* Das zweidimensionale Feld "kanten" fhrt in Verbindung mit
* einem Zhler "nachbarAnzahl" buch ber die benachbarten Gene
* jedes einzelnen Gens. Bei zwei Individuen sind pro Gen
* maximal vier Nachbarn vorhanden.
* Beispiel:
*  Falls kanten [1] = {2,4,5} ist nachbarZahl[1] = 3.
*  Gen 1 steht in beiden Chromosomen in Nachbarschaft zu den Genen
*  2, 4 und 5. Es hat also insgesamt drei unterschiedliche Nachbarn.
*/
 public Paar kreuzung( Individuum c1, Individuum c2){
      codeLaenge = c1.getCodelnge();
      if(kanten == null){
        kanten = new int[codeLaenge][4];
        nachbarAnzahl = new int[codeLaenge];
      }
      Individuum[] c = new Individuum[2];
      c[0]=c1; c[1]=c2;
      for (int k=0  ;k<codeLaenge  ; k++ ) {
             nachbarAnzahl[k]=0;
      }
      for (int k=0  ;k<codeLaenge  ; k++ ) {  //durchsuche die Gene

           for ( int nr=0 ; nr<2 ;nr++  ) {   //von jedem Individuum
            //Benachbarte Gene in c[i]
            int gen = c[nr].getGen(k);
            int posNach = (k+1)%codeLaenge;
            int genNach = c[nr].getGen(posNach);
            int posVor= k-1;
            if(posVor <0) posVor = codeLaenge-1;
            int genVor = c[nr].getGen(posVor);
            fgeEin(genNach, gen);
            fgeEin(genVor,gen);
          }
      }
      return new Paar(erx(c1),erx(c2));
    } //kreuzung
    
    /**
    *  Initialisierung  aller Gene des neuen Indivisuums mit -1
    *  Bestimmung einer Zufallsstelle
    *  Festlegung des Genwertes als Ausgangspunkt
    *  Berechnung der Route
    *    Suche des nchsten Ortes mit wenigen Verbindungen zu weiteren
    *      und Untersuchung auf Kreisfreiheit
    *    Eintragung des Ortes
    *    Weiter bis alle behandelt
    */
    public Individuum erx(Individuum i1){
      Individuum i = new Individuum(codeLaenge);
      //mit -1 belegen
      for (int k=1; k<codeLaenge; k++){
         i.setGen(k,-1);
      }
      //eine zufllige Stelle bestimmen
      int p = (int) (Math.random()*codeLaenge);  //Start mit p
      int ort = i1.getGen(p);
      // mit dem zufllig ausgewhlten Ort beginnen
      i.setGen(0,ort);
      // den Rest nach ERX auffllen
      for (int k=1; k<codeLaenge; k++){
        int naechsterOrt = mitWenigstenNachbarn(ort,i);
        i.setGen(k, naechsterOrt);
        ort=naechsterOrt;
      }
      
      return i;
    }
    
    //------------ privat und zu Testzwecken ---------------

    /**
    * prft, ob i bereits in der Nachbarschaftstabelle von nr enthalten ist
    */
    private boolean enthalten(int i, int nr){
      for(int k=0; k<nachbarAnzahl[nr]; k++){
        if (kanten[nr][k]==i){
          return true;
        } //if
      } //for
      return false;
    }
    
    /**
    * prft, ob gen im Chromosom i enthalten ist
    */
    private boolean genEnthalten(int gen, Individuum i){
      for(int k=0; k<codeLaenge; k++){
        if (i.getGen(k)==gen){
          return true;
        } //if
      } //for
      return false;
    }
    
    /**
    * fgt i in die Nachbarschaftstabelle von nr ein
    */
    private void fgeEin(int i, int nr){
      if (!enthalten(i,nr)) {
        kanten[nr][nachbarAnzahl[nr]] = i;
        nachbarAnzahl[nr]++;
        
      }
    }
    
    /**
    * sucht aus einer Nachbarschaftstabelle eines nummerierten Gens nr
    * das Gen mit den wenigsten Nachbarn. Es darf wohl nicht bereits
    * in der Rundreise (weg) aufgenommen worden sein.
    */
    public int mitWenigstenNachbarn(int nr, Individuum weg){
      int minAnz=codeLaenge;
      int min=-1; //produziert evtl. einen Fehler
      for ( int k=0 ;k < nachbarAnzahl[nr] ; k++ ) {
        int a = kanten[nr][k];
        //System.out.println(" "+a+" Nachbarn: "+nachbarAnzahl[a]+"Min akt. "+min);
        if(nachbarAnzahl[a] < minAnz && !genEnthalten(a,weg)) {
           //keine zufllige Auswahl, sondern das erste in der Liste
           minAnz = nachbarAnzahl[a];
           min = a;
        }
      }
      //Wenn alle bereits besucht worden sind, ist min == -1
      //Rckgabe des erst besten noch nicht besuchten Ortes
      if(min == -1){
        for (int k=0; k<codeLaenge; k++){
           if (!genEnthalten(k,weg)){
             min=k;
             break;
           }
        }
      }
      return min;
    }
    
    //zu Testzwecken
    public void zeigeNachbarn(){
      for ( int x=0 ; x<codeLaenge ; x++ ) {
         System.out.print(" "+x+" -> ");
           for (int y=0  ;y<nachbarAnzahl[x]  ;y++  ) {
             System.out.print(" "+kanten[x][y]+" ");
           } System.out.println();
         }
    }
    
    private int[][] kanten;
    private int[] nachbarAnzahl;
    private static int codeLaenge;
    
    
    private static DatenGA dga;
    public static void main(String[] s){
       ERX e=new ERX();
       dga = new DatenGA(codeLaenge, e,null);
       System.out.println(dga.toString());
       Individuum i1 = new Individuum(6);
       
       dga.initZufllig(i1);
       System.out.println(i1);
       Individuum i2 = new Individuum(6);
       dga.initZufllig(i2);
       System.out.println(i2);
       Paar p = dga.kreuzung(i1,i2);
       e.zeigeNachbarn();
       //System.out.println("Mit wenigsten Nachbarn von 2 "+e.mitWenigstenNachbarn(2));
       System.out.println("Die Kinder");
       System.out.println(p.getErstes());
       System.out.println(p.getZweites());
    }
}
